/******************************************************************************/
/* Global functions implementation file.                                      */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PException.h"
#include "PSynchronize.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* String resources                                                           */
/******************************************************************************/
STRINGRESOURCE_BEGIN(PErrorStrings)
	// General errors
	STRINGRESOURCE_ITEM(P_GEN_ERR_BAD_ARGUMENT,	"Invalid argument passed to function")
	STRINGRESOURCE_ITEM(P_GEN_ERR_NO_MEMORY, "Out of memory")
	STRINGRESOURCE_ITEM(P_GEN_ERR_NO_RESOURCES, "Out of resources")
	STRINGRESOURCE_ITEM(P_GEN_ERR_NO_INIT, "Can't perform the operation. Missing proper initialization")
	STRINGRESOURCE_ITEM(P_GEN_ERR_CANT_PERFORM, "Can't perform the operation")
	STRINGRESOURCE_ITEM(P_GEN_ERR_BAD_IDENTIFIER, "Invalid identifier")
	STRINGRESOURCE_ITEM(P_GEN_ERR_BAD_INDEX, "Invalid index value")
	STRINGRESOURCE_ITEM(P_GEN_ERR_UNSUPPORTED_OS, "Unsupported operative system")
	STRINGRESOURCE_ITEM(P_GEN_ERR_INSUFFICIENT_BUFFER, "The data buffer passed to a function is too small")

	// File errors
	STRINGRESOURCE_ITEM(P_FILE_ERR_FILE, "Illegal file operation")
	STRINGRESOURCE_ITEM(P_FILE_ERR_NAME_TOO_LONG, "File or directory name too long")
	STRINGRESOURCE_ITEM(P_FILE_ERR_ENTRY_NOT_FOUND, "Entry could not be found")
	STRINGRESOURCE_ITEM(P_FILE_ERR_ENTRY_EXISTS, "Entry does already exists")
	STRINGRESOURCE_ITEM(P_FILE_ERR_DISK_FULL, "There is not enough space on the disk")
	STRINGRESOURCE_ITEM(P_FILE_ERR_NOT_A_DIRECTORY, "The directory is invalid")
	STRINGRESOURCE_ITEM(P_FILE_ERR_IS_A_DIRECTORY, "File operation on a directory")
	STRINGRESOURCE_ITEM(P_FILE_ERR_SEEK, "An attempt was made to move the file pointer before the beginning of the file")
	STRINGRESOURCE_ITEM(P_FILE_ERR_EOF, "EOF was reached")
	STRINGRESOURCE_ITEM(P_FILE_ERR_LINK, "Link error in the file system")
	STRINGRESOURCE_ITEM(P_FILE_ERR_LOCKED, "File or directory is locked")
	STRINGRESOURCE_ITEM(P_FILE_ERR_FILE_CORRUPT, "File is corrupted")
	STRINGRESOURCE_ITEM(P_FILE_ERR_INVALID_ENTRY_NAME, "Entry name is invalid")
	STRINGRESOURCE_ITEM(P_FILE_ERR_BAD_ACCESS_MODE, "Invalid access mode")

	// Security errors
	STRINGRESOURCE_ITEM(P_SEC_ERR_ACCESS_DENIED, "Access denied")

	// Key errors
	STRINGRESOURCE_ITEM(P_KEY_ERR_ENTRY_EXIST, "An entry does already exist and cannot be overwritten")
	STRINGRESOURCE_ITEM(P_KEY_ERR_ENTRY_NOT_FOUND, "An entry could not be found")

	// Other errors
	STRINGRESOURCE_ITEM(P_ERR_ANY, "Any error")
	STRINGRESOURCE_ITEM(P_OK, "No error")
STRINGRESOURCE_END



/******************************************************************************/
/* Global variables                                                           */
/******************************************************************************/
static bool rndInit = false;



/******************************************************************************/
/* PlayBeep() will play the default system beep sound.                        */
/******************************************************************************/
void PlayBeep(void)
{
#if __p_os == __p_beos

	beep();

#elif __p_os == __p_windows

	MessageBeep(0xffffffff);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Random() will return a random number between 0 and max.                    */
/*                                                                            */
/* Input:  "max" is the max number you want.                                  */
/*                                                                            */
/* Output: Is a random number.                                                */
/******************************************************************************/
uint32 Random(uint32 max)
{
	// If max is 0, well, we then return 0 as the random number
	if (max == 0)
		return (0);

	if (!rndInit)
	{
		// Initialize random generator
		rndInit = true;

#if __p_os == __p_beos
		srand(real_time_clock_usecs());

#elif __p_os == __p_windows
		srand(GetTickCount());

#elif __p_os == __p_linux
implement
#endif
	}

	// This is the recommended method for getting a number between
	// 0 and max. The high bits are more random than the low bits.
    return ((uint32) ((double)(max + 1) * rand() / (RAND_MAX + 1.0)));
}



/******************************************************************************/
/* ConvertOSError() will convert an OS specific error code to a PolyKit error */
/*      code.                                                                 */
/*                                                                            */
/* Input:  "osError" is the OS error code.                                    */
/*                                                                            */
/* Output: Is the PolyKit error code.                                         */
/******************************************************************************/
uint32 ConvertOSError(uint32 osError)
{
	uint32 polyErr = P_ERR_ANY;

#if __p_os == __p_beos

	switch (osError)
	{
		//
		// General errors
		//
		case B_OK:
			polyErr = P_OK;
			break;

		case B_BAD_VALUE:
		case B_BAD_THREAD_ID:
		case B_BAD_TEAM_ID:
			polyErr = P_GEN_ERR_BAD_ARGUMENT;
			break;

		case B_NO_MEMORY:
			polyErr = P_GEN_ERR_NO_MEMORY;
			break;

		case B_NO_MORE_FDS:
		case B_NO_MORE_THREADS:
			polyErr = P_GEN_ERR_NO_RESOURCES;
			break;

		case B_NO_INIT:
			polyErr = P_GEN_ERR_NO_INIT;
			break;

		case B_NOT_ALLOWED:
		case B_BAD_THREAD_STATE:
			polyErr = P_GEN_ERR_CANT_PERFORM;
			break;

		case B_BAD_IMAGE_ID:
			polyErr = P_GEN_ERR_BAD_IDENTIFIER;
			break;

		case B_BAD_INDEX:
			polyErr = P_GEN_ERR_BAD_INDEX;
			break;

		//
		// File errors
		//
		case B_FILE_ERROR:
			polyErr = P_FILE_ERR_FILE;
			break;

		case B_NAME_TOO_LONG:
			polyErr = P_FILE_ERR_NAME_TOO_LONG;
			break;

		case B_ENTRY_NOT_FOUND:
			polyErr = P_FILE_ERR_ENTRY_NOT_FOUND;
			break;

		case B_FILE_EXISTS:
			polyErr = P_FILE_ERR_ENTRY_EXISTS;
			break;

		case B_DEVICE_FULL:
			polyErr = P_FILE_ERR_DISK_FULL;
			break;

		case B_NOT_A_DIRECTORY:
			polyErr = P_FILE_ERR_NOT_A_DIRECTORY;
			break;

		case B_LINK_LIMIT:
			polyErr = P_FILE_ERR_LINK;
			break;

		case B_BUSY:
			polyErr = P_FILE_ERR_LOCKED;
			break;

		//
		// Security errors
		//
		case B_PERMISSION_DENIED:
			polyErr = P_SEC_ERR_ACCESS_DENIED;
			break;

#ifdef PDEBUG
		default:
			printf("Error %ld: %s\n", osError, strerror(osError));
			ASSERT(false);
			break;
#endif
	}

#elif __p_os == __p_windows

	switch (osError)
	{
		//
		// General errors
		//
		case ERROR_SUCCESS:
			polyErr = P_OK;
			break;

		case ERROR_BAD_ARGUMENTS:
		case ERROR_INVALID_PARAMETER:
			polyErr = P_GEN_ERR_BAD_ARGUMENT;
			break;

		case ERROR_NOT_ENOUGH_MEMORY:
		case ERROR_OUTOFMEMORY:
			polyErr = P_GEN_ERR_NO_MEMORY;
			break;

		case ERROR_DEV_NOT_EXIST:
		case ERROR_IS_JOIN_PATH:
		case ERROR_BUSY:
		case ERROR_NO_SYSTEM_RESOURCES:
		case ERROR_NONPAGED_SYSTEM_RESOURCES:
		case ERROR_PAGED_SYSTEM_RESOURCES:
		case ERROR_RESOURCE_DISABLED:
		case ERROR_RESOURCE_NOT_PRESENT:
			polyErr = P_GEN_ERR_NO_RESOURCES;
			break;

		case ERROR_BUSY_DRIVE:
		case ERROR_IO_DEVICE:
		case ERROR_USER_MAPPED_FILE:
		case ERROR_RETRY:
		case ERROR_INVALID_STATE:
			polyErr = P_GEN_ERR_CANT_PERFORM;
			break;

		case ERROR_INVALID_HANDLE:
		case ERROR_INVALID_THREAD_ID:
		case ERROR_INVALID_TARGET_HANDLE:
			polyErr = P_GEN_ERR_BAD_IDENTIFIER;
			break;

		case ERROR_INVALID_INDEX:
			polyErr = P_GEN_ERR_BAD_INDEX;
			break;

		case ERROR_INSUFFICIENT_BUFFER:
			polyErr = P_GEN_ERR_INSUFFICIENT_BUFFER;
			break;
			
		//
		// File errors
		//
		case ERROR_TOO_MANY_OPEN_FILES:
		case ERROR_OPEN_FAILED:
			polyErr = P_FILE_ERR_FILE;
			break;

		case ERROR_BUFFER_OVERFLOW:
		case ERROR_LABEL_TOO_LONG:
			polyErr = P_FILE_ERR_NAME_TOO_LONG;
			break;

		case ERROR_FILE_NOT_FOUND:
		case ERROR_PATH_NOT_FOUND:
		case ERROR_INVALID_NAME:
			polyErr = P_FILE_ERR_ENTRY_NOT_FOUND;
			break;

		case ERROR_FILE_EXISTS:
		case ERROR_ALREADY_EXISTS:
			polyErr = P_FILE_ERR_ENTRY_EXISTS;
			break;

		case ERROR_HANDLE_DISK_FULL:
			polyErr = P_FILE_ERR_DISK_FULL;
			break;

		case ERROR_DIRECTORY:
			polyErr = P_FILE_ERR_NOT_A_DIRECTORY;
			break;

		case ERROR_HANDLE_EOF:
			polyErr = P_FILE_ERR_EOF;
			break;

		case ERROR_TOO_MANY_LINKS:
			polyErr = P_FILE_ERR_LINK;
			break;

		case ERROR_LOCK_FAILED:
		case ERROR_LOCK_VIOLATION:
			polyErr = P_FILE_ERR_LOCKED;
			break;

		case ERROR_FILE_CORRUPT:
			polyErr = P_FILE_ERR_FILE_CORRUPT;
			break;

		case ERROR_BAD_PATHNAME:
			polyErr = P_FILE_ERR_INVALID_ENTRY_NAME;
			break;

		//
		// Security errors
		//
		case ERROR_ACCESS_DENIED:
		case ERROR_SHARING_VIOLATION:
			polyErr = P_SEC_ERR_ACCESS_DENIED;
			break;

		default:
			ASSERT(false);
			break;
	}

#elif __p_os == __p_linux
implement
#endif

	return (polyErr);
}



/******************************************************************************/
/* GetErrorString() will convert an PolyKit error code to a readable string.  */
/*                                                                            */
/* Input:  "error" is the PolyKit error code.                                 */
/*                                                                            */
/* Output: Is the readable string.                                            */
/******************************************************************************/
PString GetErrorString(uint32 error)
{
	PString errStr;
	uint32 resID;

	// Add the string resources
	resID = AddStringResource(PErrorStrings);

	// Get the string
	errStr.LoadString(resID, error);

	// Remove the string resource again
	RemoveStringResource(resID);

	return (errStr);
}



/******************************************************************************/
/* GetOSVersion() will return the version of the operative system.            */
/*                                                                            */
/* Output: Is the version of the operative system.                            */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
POperativeSystem GetOSVersion(void)
{
#if __p_os == __p_beos

	return (pBeOS);

#elif __p_os == __p_windows

	OSVERSIONINFO versInfo;

	// Initialize the version info structure
	versInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	// Get the version of the operating system that is currently running
	if (!GetVersionEx(&versInfo))
		throw PSystemException(ConvertOSError(GetLastError()));

	// Use the platform id to identify the OS version
	switch (versInfo.dwPlatformId)
	{
		// Windows 95/98
		case VER_PLATFORM_WIN32_WINDOWS:

			// Check if the system is Windows 95 (minor version is 0)
			if (versInfo.dwMinorVersion == 0)
				return (pWindows95);

			// Else the system is Windows 98
			return (pWindows98);

		// Windows NT/2000
		case VER_PLATFORM_WIN32_NT:

			// Check if the system is Windows NT (version 4 or earlier)
			if (versInfo.dwMajorVersion <= 4)
				return (pWindowsNT);

			// Else the system is Windows 2000
			return (pWindows2000);

		// Unsupported OS
		default:
			throw PSystemException(P_GEN_ERR_UNSUPPORTED_OS);
	}

#elif __p_os == __p_linux
implement
#endif
}
