/******************************************************************************/
/* PString implementation file.                                               */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PSynchronize.h"
#include "PList.h"
#include "PString.h"


/******************************************************************************/
/* Structure to store resources in                                            */
/******************************************************************************/
typedef struct ResourceStruct
{
	uint32				id;
	uint32				number;
	PStringResource *	resource;
} ResourceStruct;



/******************************************************************************/
/* Global data                                                                */
/******************************************************************************/
static PMutex listLocker;
static PList<ResourceStruct *> resourceList;
static uint32 idCounter = 0;



/******************************************************************************/
/* Empty string                                                               */
/******************************************************************************/
static int32 pEmptyStringData[] = {-1, 0, 0, 0};
char *pEmptyString = (char *)&pEmptyStringData[3];



/******************************************************************************/
/* Global functions                                                           */
/******************************************************************************/

/******************************************************************************/
/* Add a resource to a list which will be searched throu when loading strings.*/
/*                                                                            */
/* Input:  "resource" is a pointer to a string resource.                      */
/*                                                                            */
/* Output: Is a unique id to use when removing the resource again.            */
/******************************************************************************/
uint32 AddStringResource(PStringResource *resource)
{
	PStringResource *strResource;
	ResourceStruct *rs;
	uint32 num = 0;

	// Start to lock the list
	if (listLocker.Lock() == pSyncOk)
	{
		try
		{
			// Allocate and fill in the structure
			rs = new ResourceStruct;
			if (rs == NULL)
				throw PMemoryException();

			// Count the number of strings
			strResource = resource;
			while ((strResource->strNum != -1) || (strResource->string != NULL))
			{
				num++;
				strResource++;
			}

			rs->id       = idCounter;
			rs->number   = num;
			rs->resource = resource;

			// Add it to the list
			resourceList.AddTail(rs);

			// Unlock it again
			listLocker.Unlock();
		}
		catch(...)
		{
			listLocker.Unlock();
			throw;
		}
	}

	// Return the unique id
	return (idCounter++);
}



/******************************************************************************/
/* Remove a resource from the resource list.                                  */
/*                                                                            */
/* Input:  "resID" is the unique id you got from the AddStringResource()      */
/*         function.                                                          */
/*                                                                            */
/* Except: PBoundsException.                                                  */
/******************************************************************************/
void RemoveStringResource(uint32 resID)
{
	int32 number, i;
	ResourceStruct *rs;

	// Start to lock the list
	if (listLocker.Lock() == pSyncOk)
	{
		try
		{
			// Try to find the item in the list
			number = resourceList.CountItems();

			for (i = 0; i < number; i++)
			{
				rs = resourceList.GetItem(i);

				if (rs->id == resID)
				{
					// We found it, so remove it
					resourceList.RemoveItem(i);

					delete rs;
					break;
				}
			}

			// Unlock it again
			listLocker.Unlock();
		}
		catch(...)
		{
			listLocker.Unlock();
			throw;
		}
	}
}



/******************************************************************************/
/* PString class                                                              */
/******************************************************************************/

/******************************************************************************/
/* Default constructor                                                        */
/******************************************************************************/
PString::PString(void)
{
	Init();
}



/******************************************************************************/
/* Constructor which sets the string to the PString given.                    */
/*                                                                            */
/* Input:  "string" is a reference to another PString object.                 */
/******************************************************************************/
PString::PString(const PString &string)
{
	if (string.GetData()->refCount >= 0)
	{
		// Copy pointer
		stringData = string.stringData;

		// Count up reference counter
		AtomicIncrement((int32 *)&GetData()->refCount);
	}
	else
	{
		Init();
		*this = string.stringData;
	}
}



/******************************************************************************/
/* Constructor which sets the string to the ascii string given.               */
/*                                                                            */
/* Input:  "string" is a pointer to a null terminated string.                 */
/******************************************************************************/
PString::PString(const char *string)
{
	uint32 strLen;

	Init();

	// If the string isn't a null pointer, count the length
	strLen = (string != NULL ? strlen(string) : 0);
	if (strLen != 0)
	{
		// Allocate the string buffer
		AllocBuffer(strLen);
		strcpy(stringData, string);
		ReleaseBuffer();
	}
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PString::~PString(void)
{
	// Free the string data
	Release();
}



/******************************************************************************/
/* GetLength() returns the length of the string in characters.                */
/*                                                                            */
/* Output: The length in characters of the string.                            */
/******************************************************************************/
int32 PString::GetLength(void) const
{
	return (GetData()->stringLen);
}



/******************************************************************************/
/* GetByteLength() returns the length of the string in bytes.                 */
/*                                                                            */
/* Output: The length in bytes of the string.                                 */
/******************************************************************************/
int32 PString::GetByteLength(void) const
{
	return (strlen(stringData));
}



/******************************************************************************/
/* GetAt() will return the byte at the index given.                           */
/*                                                                            */
/* Input:  "index" is the index in bytes into the string.                     */
/*                                                                            */
/* Output: The character at index.                                            */
/******************************************************************************/
char PString::GetAt(int32 index) const
{
	ASSERT(index >= 0);
	ASSERT(index < (int32)strlen(stringData));

	return (stringData[index]);
}



/******************************************************************************/
/* IsEmpty() will check the string to see if it's empty.                      */
/*                                                                            */
/* Output: True if the string is empty, else false.                           */
/******************************************************************************/
bool PString::IsEmpty(void) const
{
	// Empty if the length of the string is 0
	if (GetData()->stringLen == 0)
		return (true);

	// Nop, the string isn't empty
	return (false);
}



/******************************************************************************/
/* MakeEmpty() will force the string to be empty.                             */
/******************************************************************************/
void PString::MakeEmpty(void)
{
	// First release the buffer
	Release();

	// And then initialize the string
	Init();
}



/******************************************************************************/
/* MakeLower() will make the string lowercase.                                */
/******************************************************************************/
void PString::MakeLower(void)
{
	char *stringPoi;

	// Make sure we can change the string
	CopyBeforeWrite();

#if __p_os == __p_beos

	for (stringPoi = stringData; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
	{
		if ((*stringPoi & 0xc0) != 0x80)
			*stringPoi = tolower(*stringPoi);
	}

#elif __p_os == __p_windows

	for (stringPoi = stringData; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
			*stringPoi = _tolower(*stringPoi);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* MakeUpper() will make the string uppercase.                                */
/******************************************************************************/
void PString::MakeUpper(void)
{
	char *stringPoi;

	// Make sure we can change the string
	CopyBeforeWrite();

#if __p_os == __p_beos

	for (stringPoi = stringData; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
	{
		if ((*stringPoi & 0xc0) != 0x80)
			*stringPoi = toupper(*stringPoi);
	}

#elif __p_os == __p_windows

	for (stringPoi = stringData; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
			*stringPoi = _toupper(*stringPoi);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Format() will assign the APString with the text given and format it. The   */
/*      string given is the same type as with the sprintf() function.         */
/*                                                                            */
/* Input:  "formatString" is the format string.                               */
/*         "..." is the arguments to put in the format string.                */
/******************************************************************************/
void PString::Format(const char *formatString, ...)
{
	va_list argList;

	// Set argList to the first argument
	va_start(argList, formatString);

	// Format the string
	FormatIt(formatString, argList);

	// End argument list parsing
	va_end(argList);
}



/******************************************************************************/
/* Format() will assign the APString with the text given and format it. The   */
/*      string given is the same type as with the sprintf() function.         */
/*                                                                            */
/* Input:  "resID" is the resource id number.                                 */
/*         "strNum" is the string number in the string resource.              */
/*         "..." is the arguments to put in the format string.                */
/*                                                                            */
/* Except: PResourceException.                                                */
/******************************************************************************/
void PString::Format(uint32 resID, int32 strNum, ...)
{
	va_list argList;
	PString formatString;

	// Set argList to the first argument
	va_start(argList, strNum);

	// Load the string from the resource
	formatString.LoadString(resID, strNum);

	// Format the string
	FormatIt(formatString, argList);

	// End argument list parsing
	va_end(argList);
}



/******************************************************************************/
/* LoadString() will load the string resource and set the current string to   */
/*      the string resource loaded.                                           */
/*                                                                            */
/* Input:  "resID" is the resource id number.                                 */
/*         "strNum" is the resource string number.                            */
/*                                                                            */
/* Except: PResourceException.                                                */
/******************************************************************************/
void PString::LoadString(uint32 resID, int32 strNum)
{
	uint32 checkPoint;
	int32 checkNum;
	uint32 lowerBound, upperBound;
	int32 i;
	bool found = false;
	ResourceStruct *rs = NULL;

	// Search trough all the string resources
	if (listLocker.Lock() == pSyncOk)
	{
		try
		{
			// Get the resource
			for (i = 0; i < resourceList.CountItems(); i++)
			{
				rs = resourceList.GetItem(i);
				if (rs->id == resID)
				{
					found = true;
					break;
				}
			}

			// Check to see if an illegal resource id was given
			if (!found)
				throw PResourceException();

			// Make a binary search on the string number in the string resource
			// Find the middle of the string resource
			found      = false;
			lowerBound = 0;
			upperBound = rs->number - 1;

			// Do the binary search
			do
			{
				// Get the current number in the array
				checkPoint = (upperBound - lowerBound) / 2 + lowerBound;
				checkNum   = rs->resource[checkPoint].strNum;

				// Did we found it?
				if (strNum == checkNum)
					found = true;
				else
				{
					// More to check?
					if (lowerBound == upperBound)
						break;

					// Do we have to go up or down in the array?
					if (strNum > checkNum)
					{
						// Up
						lowerBound = checkPoint + 1;
					}
					else
					{
						// Down
						if (checkPoint == 0)
							break;

						upperBound = checkPoint - 1;
					}
				}
			}
			while(!found);

			// If not found, throw an exception
			if (!found)
				throw PResourceException();

			// Set the APString buffer size
			GetBuffer(strlen(rs->resource[checkPoint].string) + 1);
			strcpy(stringData, rs->resource[checkPoint].string);
			ReleaseBuffer();

			// Unlock again
			listLocker.Unlock();
		}
		catch(...)
		{
			listLocker.Unlock();
			throw;
		}
	}
}



/******************************************************************************/
/* Find() will search for a character in the string and return the position   */
/*      where it found it.                                                    */
/*                                                                            */
/* Input:  "ch" is the character to search for.                               */
/*         "startPos" is the start postion in the string to start the search. */
/*         If you want to search from the beginning of the string, give -1    */
/*         which is also the default.                                         */
/*                                                                            */
/* Output: The position in the string where the character is found or -1 if   */
/*         not found.                                                         */
/******************************************************************************/
int32 PString::Find(char ch, int32 startPos) const
{
	const char *stringPoi;

	// Skip the current character, so we don't find the same once again
	startPos++;
	ASSERT(startPos >= 0);
	ASSERT(startPos <= GetData()->stringLen);

	for (stringPoi = &stringData[startPos]; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
	{
		if (*stringPoi == ch)
			return (stringPoi - stringData);	// Found it, return the position
	}

	// If we get to this point, the character couldn't be found
	return (-1);
}



/******************************************************************************/
/* Find() will search for a substring in the string and return the position   */
/*      where it found it.                                                    */
/*                                                                            */
/* Input:  "subString" is the substring to search for.                        */
/*         "startPos" is the start postion in the string to start the search. */
/*         If you want to search from the beginning of the string, give -1    */
/*         which is also the default.                                         */
/*                                                                            */
/* Output: The position in the string where the substring is found or -1 if   */
/*         not found.                                                         */
/******************************************************************************/
int32 PString::Find(const char *subString, int32 startPos) const
{
	const char *stringPoi;
	int32 subLen;

	// Skip the current character, so we don't find the same once again
	startPos++;
	ASSERT(startPos >= 0);
	ASSERT(startPos <= GetData()->stringLen);

	// Get the length of the substring
	subLen = strlen(subString);
	if (subLen == 0)
		return (-1);	// Well, if the search string is empty, we can find it

	// Begin to search for the substring
	for (stringPoi = &stringData[startPos]; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
	{
		if (strncmp(stringPoi, subString, subLen) == 0)
			return (stringPoi - stringData);	// Found it, return the position
	}

	// If we get to this point, the character couldn't be found
	return (-1);
}



/******************************************************************************/
/* ReverseFind() will search for a character in the string starting from the  */
/*      end of it and searching backwards and then return the position where  */
/*      it found it.                                                          */
/*                                                                            */
/* Input:  "ch" is the character to search for.                               */
/*                                                                            */
/* Output: The position in the string where the character is found or -1 if   */
/*         not found.                                                         */
/******************************************************************************/
int32 PString::ReverseFind(char ch) const
{
	int32 pos;

	// Calculate the start position
	pos = GetByteLength();

	while (pos >= 0)
	{
#if __p_os == __p_beos

		// Find the previous UTF8 character
		do
		{
			pos--;

			// First 3 bytes character?
			if ((stringData[pos] & 0xf0) == 0xe0)
				break;

			// First 2 bytes character?
			if ((stringData[pos] & 0xe0) == 0xc0)
				break;

			// First 1 byte character?
			if ((stringData[pos] & 0x80) == 0x00)
				break;
		}
		while (true);

#else

		// Decrement the position counter
		pos--;

#endif

		if (stringData[pos] == ch)
			return (pos);				// Found it, return the position
	}

	// If we get to this point, the character couldn't be found
	return (-1);
}



/******************************************************************************/
/* Left() returns the left part of the current string.                        */
/*                                                                            */
/* Input:  "count" is the number of bytes you want to cut.                    */
/*                                                                            */
/* Output: The new string with the clipped context.                           */
/******************************************************************************/
PString PString::Left(int32 count) const
{
	int32 strLen;
	PString newStr;

	// Get string length
	strLen = strlen(stringData);

	// Fix out of bounds
	if (count < 0)
		count = 0;

	if (count > strLen)
		count = strLen;

	// Copy the clipped string to the new string
	AllocCopy(newStr, 0, count, 0);
	return (newStr);
}



/******************************************************************************/
/* Mid() returns a middle part of the current string.                         */
/*                                                                            */
/* Input:  "first" is the byte number from where you want to cut from.        */
/*                                                                            */
/* Output: The new string with the clipped context.                           */
/******************************************************************************/
PString PString::Mid(int32 first) const
{
	return (Mid(first, strlen(stringData) - first));
}



/******************************************************************************/
/* Mid() returns a middle part of the current string.                         */
/*                                                                            */
/* Input:  "first" is the byte number from where you want to cut from.        */
/*         "count" is the number of bytes you want to cut.                    */
/*                                                                            */
/* Output: The new string with the clipped context.                           */
/******************************************************************************/
PString PString::Mid(int32 first, int32 count) const
{
	int32 strLen;
	PString newStr;

	// Get string length
	strLen = strlen(stringData);

	// Fix out of bounds
	if (first < 0)
		first = 0;

	if (count < 0)
		count = 0;

	if ((first + count) > strLen)
		count = strLen - first;

	if (first > strLen)
		count = 0;

	// Copy the clipped string to the new string
	AllocCopy(newStr, first, count, 0);
	return (newStr);
}



/******************************************************************************/
/* Right() returns the right part of the current string.                      */
/*                                                                            */
/* Input:  "count" is the number of bytes you want to cut.                    */
/*                                                                            */
/* Output: The new string with the clipped context.                           */
/******************************************************************************/
PString PString::Right(int32 count) const
{
	int32 strLen;
	PString newStr;

	// Get string length
	strLen = strlen(stringData);

	// Fix out of bounds
	if (count < 0)
		count = 0;

	if (count > strLen)
		count = strLen;

	// Copy the clipped string to the new string
	AllocCopy(newStr, strLen - count, count, 0);
	return (newStr);
}



/******************************************************************************/
/* TrimLeft() trims leading whitespaces.                                      */
/******************************************************************************/
void PString::TrimLeft(void)
{
	char *stringPoi;
	uint32 len;

	CopyBeforeWrite();

	// Find first non-space character
	stringPoi = stringData;

	while (isspace(*stringPoi))
		stringPoi += GetCharLength(*stringPoi);

	// Fix up data and length
	len = strlen(stringData) - (stringPoi - stringData);
	memmove(stringData, stringPoi, len + 1);
	GetData()->stringLen = CountLength();
}



/******************************************************************************/
/* TrimRight() trims trailing whitespaces.                                    */
/******************************************************************************/
void PString::TrimRight(void)
{
	char *stringPoi, *last;

	CopyBeforeWrite();

	stringPoi = stringData;
	last      = NULL;

	// Find beginning of trailing spaces by starting at beginning
	while (*stringPoi != 0x00)
	{
		if (isspace(*stringPoi))
		{
			if (last == NULL)
				last = stringPoi;
		}
		else
			last = NULL;

		stringPoi += GetCharLength(*stringPoi);
	}

	if (last != NULL)
	{
		// Truncate at trailing space start
		*last = 0x00;
		GetData()->stringLen = CountLength();
	}
}



/******************************************************************************/
/* MakeNumber() creates a number string and return it.                        */
/*                                                                            */
/* Output: The new string with the number.                                    */
/******************************************************************************/
PString PString::MakeNumber(void) const
{
	PString newStr;
	uint32 pos, length, add;

	// Now insert commas
	length = GetData()->stringLen;
	pos    = 0;

	while ((pos + 3) < length)
	{
		add = (length - pos) % 3;

		if (add == 0)
			add = 3;

		newStr += Mid(pos, add) + ",";
		pos    += add;
	}

	// Add the rest
	newStr += Right(length - pos);

	return (newStr);
}



/******************************************************************************/
/* StringReplace() scans the current string for the search string given and   */
/*      replaces all instances of it. The new string is then returned.        */
/*                                                                            */
/* Input:  "search" is the string that the function has to look for.          */
/*         "replace" is the string that the "search" string is replaced with  */
/*         in the output string.                                              */
/*                                                                            */
/* Output: The string with all "search" sub-strings replaced.                 */
/******************************************************************************/
PString PString::StringReplace(const char *search, const char *replace) const
{
	int32 oldPos = -1;
	int32 pos = -1;
	int32 searchLen;
	PString out;

	// If we won't search after anything, just return the current string
	searchLen = strlen(search);
	if (searchLen == 0)
		return (stringData);

	// Begin to search after the sub-string
	while ((pos = Find(search, pos)) != -1)
	{
		// Store text in from of search string in out buffer
		if (oldPos == -1)
			out += Left(pos);
		else
			out += Mid(oldPos, pos - oldPos);

		// Add replace string to out buffer
		out += replace;

		// Remember the current position to next time
		oldPos = pos + searchLen;
	}

	// We didn't find anything, so just return the current string
	if (oldPos == -1)
		return (stringData);

	// Add remaining input to output
	out += Mid(oldPos);

	return (out);
}



/******************************************************************************/
/* CompareNoCase() will compare the string given with the current string with */
/*      a non-case sentitive compare.                                         */
/*                                                                            */
/* Input:  "string" is the string you want to compare with.                   */
/*                                                                            */
/* Output: < 0, == 0 or > 0.                                                  */
/******************************************************************************/
int32 PString::CompareNoCase(const char *string) const
{
#if __p_os == __p_beos

	return (strcasecmp(stringData, string));

#elif __p_os == __p_windows

	return (strcmpi(stringData, string));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetBuffer() will return a pointer to the character buffer. The buffer will */
/*      be minimum the size to give as argument.                              */
/*                                                                            */
/* Input:  "minBufLength" is the minimum number of character the returned     */
/*         object will have.                                                  */
/*                                                                            */
/* Output: A pointer to the string data.                                      */
/******************************************************************************/
char *PString::GetBuffer(int32 minBufLength)
{
	ASSERT(minBufLength >= 0);

	// Check to see if we need to allocate a new buffer
	if ((GetData()->refCount > 1) || (minBufLength >= GetData()->allocatedLen))
	{
		// We need to allocate a new block, but first remember the old settings
		PStringData *oldData;
		char *oldString;
		uint32 oldLength;

		oldData   = GetData();
		oldString = stringData;
		oldLength = oldData->stringLen;

		// If allocated buffer is bigger than the minimum, set it to the allocated length
		if (oldData->allocatedLen > minBufLength)
			minBufLength = oldData->allocatedLen;

		// Allocate a new memory block
		AllocBuffer(minBufLength);

		// Copy the old string into the new block of memory and set the length
		strcpy(stringData, oldString);
		GetData()->stringLen = oldLength;

		// Release the reference from the old buffer
		Release(oldData);
	}

	ASSERT(GetData()->refCount <= 1);

	// Return the pointer to the string
	return (stringData);
}



/******************************************************************************/
/* GetBufferSetLength() will return a pointer to the character buffer + set   */
/*      the length of the buffer to exactly the length given.                 */
/*                                                                            */
/* Input:  "newLength" is the extact size of the buffer you want.             */
/*                                                                            */
/* Output: Is a pointer to the character buffer of the string.                */
/******************************************************************************/
char *PString::GetBufferSetLength(int32 newLength)
{
	ASSERT(newLength >= 0);

	// Change the buffer size
	GetBuffer(newLength);

	// Set the length to the new size
	GetData()->stringLen = newLength;

	// Null terminate the string
	stringData[newLength - 1] = 0x00;

	return (stringData);
}



/******************************************************************************/
/* ReleaseBuffer() will release the control of the buffer given by the        */
/*      GetBuffer() function.                                                 */
/*                                                                            */
/* Input:  "newLength" is the new length the buffer should have.              */
/******************************************************************************/
void PString::ReleaseBuffer(int32 newLength)
{
	int32 length = newLength;

	// If needed, copy the string data to another memory block
	CopyBeforeWrite();

	// If the argument is -1, count the length of the string
	if (newLength == -1)
	{
		newLength = strlen(stringData);
		length    = CountLength();
	}

	// Debug stop: Length is bigger than the allocated one
	ASSERT(newLength < GetData()->allocatedLen);

	// Set the data to the length
	GetData()->stringLen = length;

	// Null terminate the string
	stringData[newLength] = 0x00;
}



/******************************************************************************/
/* FreeExtra() frees the extra space allocated. The new allocated block will  */
/*      have the extact length as the value returned by GetLength().          */
/******************************************************************************/
void PString::FreeExtra(void)
{
	ASSERT(GetData()->stringLen <= GetData()->allocatedLen);

	// Only allocate a new block if the lengths are different
	if (GetData()->stringLen != GetData()->allocatedLen)
	{
		PStringData *oldData;
		char *oldString;

		// Remember old data pointers
		oldData   = GetData();
		oldString = stringData;

		// Allocate new buffer
		AllocBuffer(strlen(oldString));

		// Copy the string to the new buffer
		strcpy(stringData, oldString);

		// Release the reference
		Release(oldData);

		// Count the string length
		GetData()->stringLen = CountLength();
	}
}



/******************************************************************************/
/* operator const char * will return a pointer to the string buffer for read  */
/*      only.                                                                 */
/*                                                                            */
/* Output: The pointer to the string buffer.                                  */
/******************************************************************************/
PString::operator const char * () const
{
	return (stringData);
}



/******************************************************************************/
/* operator [] will return the byte at the index given.                       */
/*                                                                            */
/* Input:  "index" is the index in bytes into the string.                     */
/*                                                                            */
/* Output: The character at index.                                            */
/******************************************************************************/
char PString::operator [] (int32 index) const
{
	return (GetAt(index));
}



/******************************************************************************/
/* operator = (PString &) will set the string to the string given.            */
/*                                                                            */
/* Input:  "string" is the new string.                                        */
/*                                                                            */
/* Output: The pointer to the string buffer.                                  */
/******************************************************************************/
const PString & PString::operator = (const PString &string)
{
	// If the string pointer is the same, do nothing because the string is
	// already assigned to the string.
	if (stringData != string.stringData)
	{
		if (string.GetData()->refCount < 0)
		{
			// Need to copy the string
			AssignCopy(string.stringData);
		}
		else
		{
			// Copy the reference
			Release();
			stringData = string.stringData;

			// Count up reference counter
			AtomicIncrement((int32 *)&GetData()->refCount);
		}
	}

	return (*this);
}



/******************************************************************************/
/* operator = (char *) will set the string to the string given.               */
/*                                                                            */
/* Input:  "string" is the new string.                                        */
/*                                                                            */
/* Output: The pointer to the string buffer.                                  */
/******************************************************************************/
const PString & PString::operator = (const char *string)
{
	if (string == NULL)
		MakeEmpty();
	else
		AssignCopy(string);

	return (*this);
}



/******************************************************************************/
/* operator + (PString &, PString &) will append the two strings given.       */
/*                                                                            */
/* Input:  "string1" is the first string.                                     */
/*         "string2" is the second string.                                    */
/*                                                                            */
/* Output: The new concatenated string.                                       */
/******************************************************************************/
PString operator + (const PString &string1, const PString &string2)
{
	PString str;

	// Append the strings
	str.ConcatCopy(string1.stringData, string1.GetByteLength(), string2.stringData, string2.GetByteLength());

	// Set the data length
	str.GetData()->stringLen = str.CountLength();

	// And return it
	return (str);
};



/******************************************************************************/
/* operator + (PString &, char *) will append the two strings given.          */
/*                                                                            */
/* Input:  "string1" is the first string.                                     */
/*         "string2" is the second string.                                    */
/*                                                                            */
/* Output: The new concatenated string.                                       */
/******************************************************************************/
PString operator + (const PString &string1, const char *string2)
{
	PString str;

	ASSERT(string2 != NULL);

	// Append the strings
	str.ConcatCopy(string1.stringData, string1.GetByteLength(), string2, strlen(string2));

	// Set the data length
	str.GetData()->stringLen = str.CountLength();

	// And return it
	return (str);
};



/******************************************************************************/
/* operator + (char *, char *) will append the two strings given.             */
/*                                                                            */
/* Input:  "string1" is the first string.                                     */
/*         "string2" is the second string.                                    */
/*                                                                            */
/* Output: The new concatenated string.                                       */
/******************************************************************************/
PString operator + (const char *string1, const PString &string2)
{
	PString str;

	ASSERT(string1 != NULL);

	// Append the strings
	str.ConcatCopy(string1, strlen(string1), string2.stringData, string2.GetByteLength());

	// Set the data length
	str.GetData()->stringLen = str.CountLength();

	// And return it
	return (str);
};



/******************************************************************************/
/* operator += (PString &) will append the string given to the current        */
/*         string.                                                            */
/*                                                                            */
/* Input:  "string" is the string to append.                                  */
/*                                                                            */
/* Output: The pointer to the string buffer.                                  */
/******************************************************************************/
const PString & PString::operator += (const PString &string)
{
	ConcatInPlace(string.stringData);
	return (*this);
}



/******************************************************************************/
/* operator += (char *) will append the string given to the current string.   */
/*                                                                            */
/* Input:  "string" is the string to append.                                  */
/*                                                                            */
/* Output: The pointer to the string buffer.                                  */
/******************************************************************************/
const PString & PString::operator += (const char *string)
{
	ASSERT(string != NULL);

	ConcatInPlace(string);
	return (*this);
}



/******************************************************************************/
/* AllocBuffer() will allocate a new memory buffer to the string + data       */
/*      structure.                                                            */
/*                                                                            */
/* Input:  "length" is the number of characters to allocate.                  */
/******************************************************************************/
void PString::AllocBuffer(int32 length)
{
	if (length == 0)
		Init();				// Makes the string empty
	else
	{
		PStringData *data;
		int memSize;

		// Allocate space to the data structure + the string
		memSize = sizeof(PStringData) + (length + 1) * sizeof(char);
		data = (PStringData *)new uint8[memSize];
		if (data == NULL)
			throw PMemoryException();

		// Initialize the structure
		data->refCount     = 1;
		data->stringLen    = length;
		data->allocatedLen = memSize - sizeof(PStringData);

		stringData = ((char *)data + sizeof(PStringData));
	}
}



/******************************************************************************/
/* Release() will count down the reference count and free the memory block if */
/*      necessary.                                                            */
/******************************************************************************/
void PString::Release(void)
{
	if (stringData != pEmptyString)
	{
		ASSERT(GetData()->refCount != 0);

		// Count down reference counter
		int32 result = AtomicDecrement((int32 *)&GetData()->refCount);

		// If there isn't any references left, delete the buffer
		if (result == 0)
		{
			delete GetData();
			Init();
		}
	}
}



/******************************************************************************/
/* Release() will count down the reference count on the data structure given  */
/*      and free the memory block if necessary.                               */
/*                                                                            */
/* Input:  "data" is a pointer to the data structure to release.              */
/******************************************************************************/
void PString::Release(PStringData *data)
{
	if (data != (PStringData *)&pEmptyStringData)
	{
		ASSERT(data->refCount != 0);

		// Count down reference counter
		int32 result = AtomicDecrement((int32 *)&data->refCount);

		// If there isn't any references left, delete the buffer
		if (result == 0)
			delete[] (uint8 *)data;
	}
}



/******************************************************************************/
/* CountLength() will count the string length in characters.                  */
/*                                                                            */
/* Output: The length in characters.                                          */
/******************************************************************************/
int32 PString::CountLength(void)
{
	int32 length;

#if __p_os == __p_beos

	char *strPoi;

	length = 0;
	strPoi = stringData;

	// We count till we meet the null terminator byte
	while (*strPoi != 0x00)
	{
		if ((*strPoi & 0xc0) != 0x80)
			length++;

		strPoi++;
	}

#elif __p_os == __p_windows

	length = strlen(stringData);

#elif __p_os == __p_linux
implement
#endif

	return (length);
}



/******************************************************************************/
/* CopyBeforeWrite() makes sure that the string buffer has it's own memory    */
/*      block.                                                                */
/******************************************************************************/
void PString::CopyBeforeWrite(void)
{
	if (GetData()->refCount > 1)
	{
		PStringData *oldData;
		char *oldString;

		// Remember old data pointers
		oldData   = GetData();
		oldString = stringData;

		// Allocate new buffer
		AllocBuffer(strlen(oldString));

		// Copy the string to the new buffer
		strcpy(stringData, oldString);

		// Release the reference
		Release(oldData);

		// Count the string length
		GetData()->stringLen = CountLength();
	}
}



/******************************************************************************/
/* AllocBeforeWrite() makes sure that there is a buffer allocated which is    */
/*      big enought to hold the new string.                                   */
/*                                                                            */
/* Input:  "strLen" is the length of the string.                              */
/******************************************************************************/
void PString::AllocBeforeWrite(int32 strLen)
{
	if ((GetData()->refCount > 1) || (GetData()->allocatedLen <= strLen))
	{
		Release();
		AllocBuffer(strLen);
	}
}



/******************************************************************************/
/* AssignCopy() assigns a new string to the this object.                      */
/*                                                                            */
/* Input:  "string" is the new string to assign with.                         */
/******************************************************************************/
void PString::AssignCopy(const char *string)
{
	// Make sure we have an allocated buffer
	AllocBeforeWrite(strlen(string));

	// Copy the string
	strcpy(stringData, string);

	// Set the data length
	GetData()->stringLen = CountLength();
}



/******************************************************************************/
/* AllocCopy() will copy a clipped area from the this object to a new string. */
/*                                                                            */
/* Input:  "dest" is a reference to where to copy the clipped data.           */
/*         "copyIndex" is where to start the copy.                            */
/*         "copyLen" is the number of bytes to copy.                          */
/*         "extraLen" is extra number of bytes to allocate.                   */
/******************************************************************************/
void PString::AllocCopy(PString &dest, int32 copyIndex, int32 copyLen, int32 extraLen) const
{
	int32 newLen;

	// Calculate the new length
	newLen = copyLen + extraLen;
	if (newLen == 0)
		dest.Init();
	else
	{
		// Allocate new buffer and copy the data
		dest.AllocBuffer(newLen);
		memcpy(dest.stringData, stringData + copyIndex, copyLen);
		dest.stringData[newLen] = 0x00;

		// Set the string length
		dest.GetData()->stringLen = dest.CountLength();
	}
}



/******************************************************************************/
/* ConcatInPlace() appends a string to the this object.                       */
/*                                                                            */
/* Input:  "string" is the string to append.                                  */
/******************************************************************************/
void PString::ConcatInPlace(const char *string)
{
	int32 strLen;

	// Get the string length
	strLen = strlen(string);

	if (strLen == 0)
		return;			// Append an empty string is a no-operation

	// Check to see if we need to allocate a new buffer
	if ((GetData()->refCount > 1) || ((int32)strlen(stringData) + strLen >= GetData()->allocatedLen))
	{
		// We have to grow the buffer
		PStringData *oldData = GetData();
		ConcatCopy(stringData, strlen(stringData), string, strLen);
		Release(oldData);
	}
	else
	{
		// Just append the string, the buffer is big enough
		strcat(stringData, string);
	}

	// Set the data length
	GetData()->stringLen = CountLength();
}



/******************************************************************************/
/* ConcatCopy() append two strings and store it in the this object.           */
/*                                                                            */
/* Input:  "sourceString" is the start string.                                */
/*         "sourceLen" is the length of the start string.                     */
/*         "appendString" is the string to append.                            */
/*         "appendLen" is the length of the append string.                    */
/******************************************************************************/
void PString::ConcatCopy(const char *sourceString, int32 sourceLen, const char *appendString, int32 appendLen)
{
	int32 newLen;

	// Calculate the new length
	newLen = sourceLen + appendLen;

	if (newLen != 0)
	{
		// Allocate new buffer and append the strings
		AllocBuffer(newLen);
		memcpy(stringData, sourceString, sourceLen);
		memcpy(stringData + sourceLen, appendString, appendLen);

		// Null terminate the new string
		stringData[newLen] = 0x00;
	}
}



/******************************************************************************/
/* FormatIt() is the real sprintf like format routine.                        */
/*                                                                            */
/* Input:  "formatString" is the format string.                               */
/*         "argList" is a pointer to the arguments.                           */
/******************************************************************************/
void PString::FormatIt(const char *formatString, va_list argList)
{
	int32 maxLen, itemLen;
	int32 width, precision;
	const char *stringPoi;
	va_list argListSave = argList;

	// Make a guess at the maximum length of the string
	maxLen = 0;

	for (stringPoi = formatString; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
	{
		// Handle '%' character, but watch out for '%%'
		if ((*stringPoi != '%') || (*(stringPoi += GetCharLength(*stringPoi)) == '%'))
		{
			maxLen += GetCharLength(*stringPoi);
			continue;
		}

		// Found a '%' character, so parse it like the sprintf function will do it
		// First handle the width argument
		itemLen = 0;
		width   = 0;

		for (; *stringPoi != 0x00; stringPoi += GetCharLength(*stringPoi))
		{
			// Check for valid flags
			if (*stringPoi == '#')
				maxLen += 2;		// For '0x'
			else
				if (*stringPoi == '*')
					width = va_arg(argList, int);
				else
					if ((*stringPoi == '-') || (*stringPoi == '+') ||
						(*stringPoi == '0') || (*stringPoi == ' '))
						continue;
					else
						break;
		}

		// Get width and skip it
		if (width == 0)
		{
			// Width indicated by a number in the string
			width = atoi(stringPoi);

			// Skip the number in the string
			while ((*stringPoi != 0x00) && (isdigit(*stringPoi)))
				stringPoi += GetCharLength(*stringPoi);
		}
		ASSERT(width >= 0);

		// Handle the precision argument
		precision = 0;
		if (*stringPoi == '.')
		{
			// Skip past '.' separator (width.precision)
			stringPoi += GetCharLength(*stringPoi);

			// Get the precision and skip it
			if (*stringPoi == '*')
			{
				precision = va_arg(argList, int);
				stringPoi += GetCharLength(*stringPoi);
			}
			else
			{
				// Precision indicated by a number in the string
				precision = atoi(stringPoi);

				// Skip the number in the string
				while ((*stringPoi != 0x00) && (isdigit(*stringPoi)))
					stringPoi += GetCharLength(*stringPoi);
			}
			ASSERT(precision >= 0);
		}

		// Should be on type modifier or specifier
		switch (*stringPoi)
		{
			case 'h':
			case 'l':
			case 'F':
			case 'N':
			case 'L':
				stringPoi += GetCharLength(*stringPoi);
				break;
		}

		// Now should be on specifier
		switch (*stringPoi)
		{
			// Single characters
			case 'c':
			{
				int tempArg;

				itemLen = 4;					// Max UTF8 character length
				tempArg = va_arg(argList, int);	// Skip argument
				break;
			}

			// Strings
			case 's':
				itemLen = strlen(va_arg(argList, const char *));
				itemLen = max(1, itemLen);
				break;
		}

		// Adjust itemLen for strings
		if (itemLen != 0)
		{
			itemLen = max(itemLen, width);
			if (precision != 0)
				itemLen = min(itemLen, precision);
		}
		else
		{
			switch (*stringPoi)
			{
				// Integers
				case 'd':
				case 'i':
				case 'u':
				case 'x':
				case 'X':
				case 'o':
				{
					int tempArg;

					tempArg = va_arg(argList, int);
					itemLen = max(32, width + precision);
					break;
				}

				// Floating point
				case 'e':
				case 'E':
				case 'f':
				case 'g':
				case 'G':
				{
					double tempArg;

					tempArg = va_arg(argList, double);
					itemLen = max(128, width + precision);
					break;
				}

				// Pointer
				case 'p':
				{
					void *tempArg;

					tempArg = va_arg(argList, void *);
					itemLen = max(32, width + precision);
					break;
				}

				// No output
				case 'n':
				{
					int *tempArg;

					tempArg = va_arg(argList, int *);
					break;
				}

				// Unknown formatting option
				default:
					ASSERT(false);
			}
		}

		// Adjust maxLen for output itemLen
		maxLen += itemLen;
	}

	// Create the new string
	GetBuffer(maxLen);
	VERIFY(vsprintf(stringData, formatString, argListSave) <= GetData()->allocatedLen);
	ReleaseBuffer();

	va_end(argListSave);
}
