/******************************************************************************/
/* PSynchronize implementation file.                                          */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PSynchronize.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* Global functions                                                           */
/******************************************************************************/

/******************************************************************************/
/* AtomicIncrement() will increment the variable you give by one and return   */
/*      the result. It makes sure that only one thread can access the         */
/*      variable at the time.                                                 */
/*                                                                            */
/* Input:  "variable" is a pointer to the variable you want to increment.     */
/*                                                                            */
/* Output: Is the result after the increment.                                 */
/******************************************************************************/
int32 AtomicIncrement(int32 *variable)
{
#if __p_os == __p_beos

	int32 result;

	result = atomic_add(variable, 1);
	return (result + 1);

#elif __p_os == __p_windows

	return (InterlockedIncrement(variable));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* AtomicDecrement() will decrement the variable you give by one and return   */
/*      the result. It makes sure that only one thread can access the         */
/*      variable at the time.                                                 */
/*                                                                            */
/* Input:  "variable" is a pointer to the variable you want to decrement.     */
/*                                                                            */
/* Output: Is the result after the decrement.                                 */
/******************************************************************************/
int32 AtomicDecrement(int32 *variable)
{
#if __p_os == __p_beos

	int32 result;

	result = atomic_add(variable, -1);
	return (result - 1);

#elif __p_os == __p_windows

	return (InterlockedDecrement(variable));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* MultipleObjectsWait() will wait on multiple synchronize objects. You can   */
/*      select between you want to wait on all the objects or only on one     */
/*      of them.                                                              */
/*                                                                            */
/* Input:  "objects" is a pointer to an array with pointers to the objects.   */
/*         "count" is the number of objects to wait on.                       */
/*         "waitAll" selects between a "wait on all" or "wait on one".        */
/*         "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: Is the array number of the object you got, or one of the errors.   */
/******************************************************************************/
int32 MultipleObjectsWait(PSync **objects, int32 count, bool waitAll, uint32 timeout)
{
	// Check arguments
	ASSERT(objects != NULL);
	ASSERT(count > 0);

#if __p_os == __p_beos

	PSyncError err;
	int32 i;
	bigtime_t time;

	if (waitAll)
	{
		// We need to wait on all the objects
		PSync *waitObj;
		int32 waitNum;
		int32 j;
		bool gotAll;

		// We just start to wait on the first object
		waitObj = objects[0];
		waitNum = 0;

		do
		{
			// Wait on the object
			time = system_time();
			err = waitObj->Lock(timeout);
			if (err != pSyncOk)
				return (err);

			// We got the object, can we get the rest
			gotAll = true;
			for (i = 0; i < count; i++)
			{
				if (i != waitNum)
				{
					// Try with no timeout
					err = objects[i]->Lock(0);
					if (err == pSyncError)
					{
						// Unlock the objects we have locked
						for (j = i - 1; j >= 0; j--)
							objects[j]->Unlock();

						return (err);
					}

					if (err == pSyncTimeout)
					{
						// We didn't get it, so unlock the objects
						// we already have locked and wait on this
						// object
						for (j = i - 1; j >= 0; j--)
							objects[j]->Unlock();

						waitObj = objects[i];
						waitNum = i;
						gotAll  = false;

						// Calculate rest timeout
						if (timeout != PSYNC_INFINITE)
							timeout -= ((system_time() - time) / 1000);
						break;
					}
				}
			}
		}
		while (!gotAll);

		return (0);		// Just return the first object index
	}
	else
	{
		// We only need to wait on one object
		for (;;)
		{
			// Get the current system time
			time = system_time();

			// Check to see if we can lock one of the objects
			for (i = 0; i < count; i++)
			{
				err = objects[i]->Lock(0);
				if (err == pSyncOk)
					return (i);			// Yeah, return the index

				if (err == pSyncError)
					return (err);
			}

			// Nop, we couldn't lock any of the objects, so wait
			// max 3 milliseconds and try again
			snooze(min(3000, timeout * 1000));

			// Check for timeout
			if (timeout != PSYNC_INFINITE)
			{
				time = (system_time() - time) / 1000;
				if (time >= timeout)
					return (pSyncTimeout);

				timeout -= time;
			}
		}

		return (pSyncError);
	}

#elif __p_os == __p_windows

	HANDLE	*handles;
	DWORD	res;

	//
	// Create a dynamic array of internal object handles
	//

	// Allocate an array of for object handles
	handles = new HANDLE[count];
	if (handles == NULL)
		throw PMemoryException();

	// Store the all object handles into the array of object handles
	for (int32 i = 0; i < count; i++)
		handles[i] = objects[i]->GetHandle();

	//
	// Wait for the specified objects
	//
	res = WaitForMultipleObjects(count, handles, waitAll, timeout == PSYNC_INFINITE ? INFINITE : timeout);

	// Check if the return value is between WAIT_OBJECT_0 to (WAIT_OBJECT_0 + count - 1)
	if ((res >=  WAIT_OBJECT_0) && (res < (WAIT_OBJECT_0 + count)))
	{
		// If waitAll is true, the return value indicates that the state of
		// all specified objects is signaled.
		if (waitAll)
			return (0);	// Just return the first object index

		// Else, the return value minus WAIT_OBJECT_0 indicates the handles
		// array index of the object that satisfied the wait. 
		else
			return (res - WAIT_OBJECT_0);
	}
	// Else check if the time-out interval elapsed
	else if (res == WAIT_TIMEOUT)
		return (pSyncTimeout);

	// The waiting was abondoned or an error orcurred
	return (pSyncError);

	// Delete the array of object handles
	delete[] handles;

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* PSync class                                                                */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/
PSync::PSync(void)
{
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PSync::~PSync(void)
{
}



/******************************************************************************/
/* PSemaphore class                                                           */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "count" is the start count of the semaphore.                       */
/*         "lockSameThread" set this if you want the thread that already got  */
/*         the semaphore to be locked when it calls this function again.      */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
PSemaphore::PSemaphore(uint32 count, bool lockSameThread)
{
#if __p_os == __p_beos

	// Create the semaphore
	semID = create_sem(count, "Some PSemaphore");
	if (semID < B_NO_ERROR)
		throw PSystemException(ConvertOSError(semID));

	// Initialize owner variables
	owner      = 0;
	ownerCount = 0;
	lockThread = lockSameThread;

#elif __p_os == __p_windows

	// Create the semaphore
	if (!(semaphore = CreateSemaphore(NULL, count, LONG_MAX, NULL)))
		throw PSystemException(ConvertOSError(GetLastError()));

	// Initialize owner variables
	owner      = 0;
	lockThread = lockSameThread;

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PSemaphore::~PSemaphore(void)
{
#if __p_os == __p_beos

	// Delete the semaphore
	VERIFY(delete_sem(semID) == B_NO_ERROR);

#elif __p_os == __p_windows

	// Close the semaphore
	VERIFY(CloseHandle(semaphore));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Lock() will wait on the object and first return when it got it or a        */
/*      timeout has occurred.                                                 */
/*                                                                            */
/* Input:  "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: pSyncOk if you got the object, else one of the PSyncError codes.   */
/******************************************************************************/
PSyncError PSemaphore::Lock(uint32 timeout)
{
#if __p_os == __p_beos

	status_t err;
	thread_id callThread;

	// Get the current thread id
	callThread = find_thread(NULL);

	// Check to see if the calling thread already have the semaphore
	if (!lockThread && (callThread == owner))
	{
		ownerCount++;
		return (pSyncOk);
	}

	// Try to acquire the semaphore
	if (timeout == PSYNC_INFINITE)
		err = acquire_sem(semID);
	else
		err = acquire_sem_etc(semID, 1, B_TIMEOUT, timeout * 1000);

	if (err == B_NO_ERROR)
	{
		owner      = callThread;
		ownerCount = 1;

		return (pSyncOk);
	}

	if ((err == B_TIMED_OUT) || ((err == B_WOULD_BLOCK) && (timeout == 0)))
		return (pSyncTimeout);

	return (pSyncError);

#elif __p_os == __p_windows

	DWORD err;
	DWORD callThread;

	// Get the current thread id
	callThread = GetCurrentThreadId();

	// Lock the thread if it is already the owner
	if (lockThread && (callThread == owner))
	{
		Sleep(timeout == PSYNC_INFINITE ? INFINITE : timeout);
		return (pSyncTimeout);
	}

	// Wait for the semaphore to become available
	err = WaitForSingleObject(semaphore, timeout == PSYNC_INFINITE ? INFINITE : timeout);

	// Return error code
	switch(err)
	{
		// The state of the semaphore is signaled
		case WAIT_OBJECT_0:
			owner = callThread;
			return (pSyncOk);

		// The time-out interval elapsed, and the semaphore's state is nonsignaled
		case WAIT_TIMEOUT:
			return (pSyncTimeout);

		// The wait function failed
		default:
			return (pSyncError);
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Unlock() will release the object.                                          */
/*                                                                            */
/* Output: pSyncOk if you the object got released else pSyncError.            */
/******************************************************************************/
PSyncError PSemaphore::Unlock(void)
{
#if __p_os == __p_beos

	// Release the semaphore
	status_t err;

	// Decrement the owner counter
	ownerCount--;

	if ((ownerCount <= 0) || (lockThread))
	{
		owner      = 0;
		ownerCount = 0;

		err = release_sem(semID);
		if (err != B_NO_ERROR)
			return (pSyncError);
	}

	return (pSyncOk);

#elif __p_os == __p_windows

	// Release the semaphore
	owner = 0;

	if (!ReleaseSemaphore(semaphore, 1, NULL))
		return (pSyncError);

	return (pSyncOk);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* PMutex class                                                               */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "lockSameThread" set this if you want the thread that already got  */
/*         the semaphore to be locked when it calls this function again.      */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
PMutex::PMutex(bool lockSameThread)
{
#if __p_os == __p_beos

	// Create the semaphore
	semID = create_sem(1, "Some PMutex");
	if (semID < B_NO_ERROR)
		throw PSystemException(ConvertOSError(semID));

	// Initialize owner variables
	owner      = 0;
	ownerCount = 0;
	lockThread = lockSameThread;

#elif __p_os == __p_windows

	// Create the mutex
	if (!(mutex = CreateMutex(NULL, FALSE, NULL)))
		throw PSystemException(ConvertOSError(GetLastError()));

	// Initialize owner variables
	owner      = 0;
	lockThread = lockSameThread;

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PMutex::~PMutex(void)
{
#if __p_os == __p_beos

	// Delete the semaphore
	VERIFY(delete_sem(semID) == B_NO_ERROR);

#elif __p_os == __p_windows

	// Close the mutex
	VERIFY(CloseHandle(mutex));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Lock() will wait on the object and first return when it got it or a        */
/*      timeout has occurred.                                                 */
/*                                                                            */
/* Input:  "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: pSyncOk if you got the object, else one of the PSyncError codes.   */
/******************************************************************************/
PSyncError PMutex::Lock(uint32 timeout)
{
#if __p_os == __p_beos

	status_t err;
	thread_id callThread;

	// Get the current thread id
	callThread = find_thread(NULL);

	// Check to see if the calling thread already have the semaphore
	if (!lockThread && (callThread == owner))
	{
		ownerCount++;
		return (pSyncOk);
	}

	// Try to acquire the semaphore
	if (timeout == PSYNC_INFINITE)
		err = acquire_sem(semID);
	else
		err = acquire_sem_etc(semID, 1, B_TIMEOUT, timeout * 1000);

	if (err == B_NO_ERROR)
	{
		owner      = callThread;
		ownerCount = 1;

		return (pSyncOk);
	}

	if ((err == B_TIMED_OUT) || ((err == B_WOULD_BLOCK) && (timeout == 0)))
		return (pSyncTimeout);

	return (pSyncError);

#elif __p_os == __p_windows

	DWORD err;
	DWORD callThread;

	// Get the current thread id
	callThread = GetCurrentThreadId();

	// Lock the thread if it is already the owner
	if (lockThread && (callThread == owner))
	{
		Sleep(timeout == PSYNC_INFINITE ? INFINITE : timeout);
		return (pSyncTimeout);
	}

	// Wait for the mutex to become available
	err = WaitForSingleObject(mutex, timeout == PSYNC_INFINITE ? INFINITE : timeout);

	// Return error code
	switch (err)
	{
		// The state of the mutex is signaled
		case WAIT_OBJECT_0:
			owner = callThread;
			return (pSyncOk);

		// The time-out interval elapsed, and the mutex's state is nonsignaled
		case WAIT_TIMEOUT:
			return (pSyncTimeout);

		// The wait function failed
		default:
			return (pSyncError);
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Unlock() will release the object.                                          */
/*                                                                            */
/* Output: pSyncOk if you the object got released else pSyncError.            */
/******************************************************************************/
PSyncError PMutex::Unlock(void)
{
#if __p_os == __p_beos

	// Release the semaphore
	status_t err;

	// Decrement the owner counter
	ownerCount--;

	if ((ownerCount <= 0) || (lockThread))
	{
		owner      = 0;
		ownerCount = 0;

		err = release_sem(semID);
		if (err != B_NO_ERROR)
			return (pSyncError);
	}

	return (pSyncOk);

#elif __p_os == __p_windows

	// Release the mutex
	owner = 0;

	if (!ReleaseMutex(mutex))
		return (pSyncError);

	return (pSyncOk);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* PEvent class                                                               */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "manualReset" indicates if you want to create a manual or          */
/*         automatic reset event.                                             */
/*         "initialState" indicates what state you want the event to be       */
/*         created in.                                                        */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
PEvent::PEvent(bool manualReset, bool initialState)
{
#if __p_os == __p_beos

	// Remember the arguments
	state         = initialState;
	manual        = manualReset;
	releaseThread = false;

	// Create the semaphore
	if (state)
		semID = create_sem(1, "Some PEvent");
	else
		semID = create_sem(0, "Some PEvent");

	if (semID < B_NO_ERROR)
		throw PSystemException(ConvertOSError(semID));

#elif __p_os == __p_windows

	// Create the event
	if (!(event = CreateEvent(NULL, manualReset, initialState, NULL)))
		throw PSystemException(ConvertOSError(GetLastError()));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PEvent::~PEvent(void)
{
#if __p_os == __p_beos

	// Delete the semaphore
	VERIFY(delete_sem(semID) == B_NO_ERROR);

#elif __p_os == __p_windows

	// Close the event
	VERIFY(CloseHandle(event));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Lock() will wait on the object and first return when it got it or a        */
/*      timeout has occurred.                                                 */
/*                                                                            */
/* Input:  "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: pSyncOk if you got the object, else one of the PSyncError codes.   */
/******************************************************************************/
PSyncError PEvent::Lock(uint32 timeout)
{
#if __p_os == __p_beos

	status_t err;
	bool releaseMe;

	// First check to see if the event has been set
	if (state)
		return (pSyncOk);		// It has, so just return

	do
	{
		// Try to acquire the semaphore
		if (timeout == PSYNC_INFINITE)
			err = acquire_sem(semID);
		else
			err = acquire_sem_etc(semID, 1, B_TIMEOUT, timeout * 1000);

		if ((err == B_TIMED_OUT) || ((err == B_WOULD_BLOCK) && (timeout == 0)))
			return (pSyncTimeout);

		if (err != B_NO_ERROR)
			return (pSyncError);

		releaseMe     = releaseThread;
		releaseThread = false;

		// Release the semaphore again so other waiting threads will
		// be signaled too
		err = release_sem(semID);
		if (err != B_NO_ERROR)
			return (pSyncError);
	}
	while ((!state) && (!releaseMe));

	return (pSyncOk);

#elif __p_os == __p_windows

	DWORD err;

	// Wait for the event to become signaled
	err = WaitForSingleObject(event, timeout == PSYNC_INFINITE ? INFINITE : timeout);

	// Return error code
	switch (err)
	{
		// The state of the event is signaled
		case WAIT_OBJECT_0:
			return (pSyncOk);

		// The time-out interval elapsed, and the event's state is nonsignaled
		case WAIT_TIMEOUT:
			return (pSyncTimeout);

		// The wait function failed
		default:
			return (pSyncError);
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Unlock() will release the object.                                          */
/*                                                                            */
/* Output: pSyncOk if you the object got released else pSyncError.            */
/******************************************************************************/
PSyncError PEvent::Unlock(void)
{
	return (pSyncError);
}



/******************************************************************************/
/* SetEvent() will set the event.                                             */
/*                                                                            */
/* Output: pSyncOk for success, pSyncError for an error.                      */
/******************************************************************************/
PSyncError PEvent::SetEvent(void)
{
#if __p_os == __p_beos

	status_t err;

	// First check to see if the event already is set
	if (state)
		return (pSyncOk);		// It is, so just return

	// Set the event
	if (manual)
		state = true;
	else
		releaseThread = true;

	err = release_sem(semID);
	if (err == B_NO_ERROR)
	{
		if (!manual)
		{
			// Automatic event, wait a little bit so we get a task-switch
			// and then reset the event again
			snooze(1000);
			acquire_sem(semID);
		}

		return (pSyncOk);
	}

	return (pSyncError);

#elif __p_os == __p_windows

	// Set the event
	return (::SetEvent(event) ? pSyncOk : pSyncError);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* ResetEvent() will clear the event.                                         */
/*                                                                            */
/* Output: pSyncOk for success, pSyncError for an error.                      */
/******************************************************************************/
PSyncError PEvent::ResetEvent(void)
{
#if __p_os == __p_beos

	status_t err;

	// First check to see if the event already is cleared
	if (!state)
		return (pSyncOk);		// It is, so just return

	// Clear the event
	err = acquire_sem(semID);
	if (err == B_NO_ERROR)
	{
		state = false;
		return (pSyncOk);
	}

	return (pSyncError);

#elif __p_os == __p_windows

	// Reset the event
	return (::ResetEvent(event) ? pSyncOk : pSyncError);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* PSWMRLocker class                                                          */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
PSWMRLocker::PSWMRLocker(void)
{
	// Initialize member variables
	mutexNoWriter  = NULL;
	eventNoReaders = NULL;
	semNumReaders  = NULL;

	try
	{
		// This mutex guards access to the other objects
		// managed by this class and also indicates whether
		// any writer threads are writing.
		mutexNoWriter = new PMutex(true);
		if (mutexNoWriter == NULL)
			throw PMemoryException();

		// Create the manual-reset event that is signaled when
		// no reader threads are reading.
		eventNoReaders = new PEvent(true, true);
		if (eventNoReaders == NULL)
			throw PMemoryException();

		// Initialize the semaphore that indicates the number
		// of reader threads that are reading.
		semNumReaders = new PSemaphore(0, true);
		if (semNumReaders == NULL)
			throw PMemoryException();
	}
	catch(PException *)
	{
		delete semNumReaders;
		delete eventNoReaders;
		delete mutexNoWriter;
		throw;
	}
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PSWMRLocker::~PSWMRLocker(void)
{
	delete semNumReaders;
	delete eventNoReaders;
	delete mutexNoWriter;
}



/******************************************************************************/
/* WaitToWrite() will wait on the object is ready to gain write access. It    */
/*      will first return when it's safe to write or a timeout has occurred.  */
/*                                                                            */
/* Input:  "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: pSyncOk if you got the object, else one of the PSyncError codes.   */
/******************************************************************************/
PSyncError PSWMRLocker::WaitToWrite(uint32 timeout)
{
	PSync *objects[2];
	int32 index;

	// We can write if the following are true:
	//
	// 1. The mutex guard is available and no other threads are writing.
	// 2. No threads are reading.
	objects[0] = mutexNoWriter;
	objects[1] = eventNoReaders;

	index = MultipleObjectsWait(objects, 2, true, timeout);
	if (index < 0)
		return ((PSyncError)index);

	return (pSyncOk);
}



/******************************************************************************/
/* DoneWriting() will release the write access, so other threads can read or  */
/*      write.                                                                */
/*                                                                            */
/* Output: pSyncOk for success, pSyncError for an error.                      */
/******************************************************************************/
PSyncError PSWMRLocker::DoneWriting(void)
{
	// Presumably, a writer thread calling this function has
	// successfully called WaitToWrite(). This means that we
	// do not have to wait on any synchronization objects
	// here, because the writer already owns the mutex.
	return (mutexNoWriter->Unlock());
}



/******************************************************************************/
/* WaitToRead() will wait on the object is ready to gain read access. It      */
/*      will first return when it's safe to read or a timeout has occurred.   */
/*                                                                            */
/* Input:  "timeout" is the timeout value in milliseconds.                    */
/*                                                                            */
/* Output: pSyncOk if you got the object, else one of the PSyncError codes.   */
/******************************************************************************/
PSyncError PSWMRLocker::WaitToRead(uint32 timeout)
{
	PSyncError err;

	// We can read if the mutex guard is available and no threads are writing
	err = mutexNoWriter->Lock(timeout);

	if (err == pSyncOk)
	{
		// This thread can read from the shared data
		//
		// Increment the number of reader threads
		err = semNumReaders->Unlock();
		if (err == pSyncOk)
		{
			// Set the event to indicate that there is a reader thread
			err = eventNoReaders->ResetEvent();
			if (err != pSyncOk)
				semNumReaders->Lock();
		}

		// Allow other writer/reader threads to use
		// the synchronization object
		mutexNoWriter->Unlock();
	}

	return (err);
}



/******************************************************************************/
/* DoneReading() will release the read access, so other threads can write.    */
/*                                                                            */
/* Output: pSyncOk for success, pSyncError for an error.                      */
/******************************************************************************/
PSyncError PSWMRLocker::DoneReading(void)
{
	PSync *objects[2];
	int32 index;
	bool lastReader;

	// We can stop reading if the mutex guard is available,
	// but when we stop reading we must also decrement the
	// number of reader threads.
	objects[0] = mutexNoWriter;
	objects[1] = semNumReaders;

	index = MultipleObjectsWait(objects, 2, true);
	if (index < 0)
		return ((PSyncError)index);

	lastReader = (semNumReaders->Lock(0) == pSyncTimeout);
	if (lastReader)
	{
		// If this is the last reader thread, set the event
		// to reflect this
		eventNoReaders->SetEvent();
	}
	else
	{
		// If this is NOT the last reader thread, we successfully
		// waited on the semaphore. We must release the semaphore
		// so that the count accurately reflects the number
		// of reader threads.
		semNumReaders->Unlock();
	}

	// Allow other writer/reader threads to use
	// the synchronization object
	mutexNoWriter->Unlock();

	return (pSyncOk);
}
